#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this class like this:

    >>> from waste_flow.analysis import waste_ana
    >>> print(waste_ana.summary_recovered)
"""

# Built-in modules #

# Internal modules #
from waste_flow            import cache_dir
from waste_flow.treatment  import waste_trt
from waste_flow.generation import waste_gen
from waste_flow.common     import trt_names, wastes_created, wet_coefs
from waste_flow.spreading  import spread

# First party modules #
from plumbing.cache import property_cached, property_pickled_at

# Third party modules #
import pandas

###############################################################################
class WasteAnalysis:
    """
    Here we combine the waste generation and waste treatment datasets and
    perform a custom analysis by breaking down the waste categories into
    new categories and calculating which proportion of waste ends up in which
    treatment category per activity sector.
    """

    # ----------------------------- Preparing ------------------------------- #
    @property
    def combined(self):
        """
        Starting from the waste generation dataframe, add the waste
        treatment data by performing a join on country, year and
        waste category.
        Now we have both the activity sector and the waste destinations
        combined.
        """
        # Load #
        gen = waste_gen.long_format
        trt = waste_trt.normalized
        # Join #
        df = gen.left_join(trt, on = ['country', 'year', 'waste'])
        # Return #
        return df

    @property
    def kilograms(self):
        """
        Multiply the waste total by the fraction that goes to each
        treatment destination. Also convert to kilograms which is
        the SI unit.
        """
        # Load #
        df = self.combined
        # Process #
        df['kg'] = df['tonnes'] * df['frac'] * 1000
        # Drop #
        df = df.drop(columns=['tonnes', 'frac'])
        # Return #
        return df

    # ----------------------------- Processing ------------------------------ #
    @property_pickled_at('cache_path')
    def spread_waste(self):
        """
        The eurostat original waste categories are spread into new custom
        waste categories. These categories have modified names and are divided
        into "household" categories and "industrial" categories.

        In particular, we remove the municipal waste (W101) category and
        distribute it in other categories according to specific breakdown
        proportions.
        All these proportions are described in "waste_spreading.csv".

        The municipal waste (W101) should be called as such because otherwise
        it can be confused with the "household" nace sector.
        """
        # Load dataframe #
        df = self.kilograms
        # Group #
        groups = df.groupby(['country', 'year', 'wst_oper', 'nace_r2'])
        # Apply function #
        df = groups.apply(lambda x: self.spreader(*x.name, x))
        # Return #
        return df

    @staticmethod
    def spreader(country, year, treatment, nace, subdf):
        # Load the coeffs #
        coeffs  = spread.by_nace[nace]
        # Multiply #
        wastes  = subdf.set_index('waste')['kg']
        product = coeffs * wastes
        # Replace some NaNs with zeros #
        product[coeffs == 0.0] = 0.0
        # But otherwise keep NaNs #
        summed  = product.sum(axis=1, skipna=False)
        # Return #
        return summed

    @property_cached
    def dry_mass(self):
        """
        Convert all the 'wet tonnes' values to their dry weight
        equivalent.
        """
        # Compute dry coefficients #
        dry_coef = 1 - wet_coefs
        # Multiply for dry mass #
        result = self.spread_waste * dry_coef
        # Return #
        return result

    @property
    def dry_long(self):
        """Same as above but in the long format."""
        # Load #
        df = self.dry_mass
        # Unpivot #
        df = df.reset_index()
        df = df.melt(id_vars    = ['country', 'year', 'nace_r2', 'wst_oper'],
                     value_name = 'kg_dry',
                     var_name   = 'waste')
        # Return #
        return df

    @property
    def collapse_ind(self):
        """
        Sum all nace sectors together, keeping only the <EP_HH>
        one as separate (and thus not summed).
        """
        # Load #
        df = self.dry_long
        # Split the dataset in two #
        selector = df['nace_r2'] == 'EP_HH'
        house = df[selector]
        indus = df[~selector]
        # Reset index #
        indus = indus.reset_index(drop=True)
        house = house.reset_index(drop=True)
        # Sum the industrial #
        groups = indus.groupby(['country', 'year', 'wst_oper', 'waste'])
        summed = groups.aggregate({'kg_dry': 'sum'})
        summed = summed.reset_index()
        # Create a new artificial nace that is the sum of the others #
        summed.insert(2, "nace_r2", "all_indus")
        # Put them back together #
        df = pandas.concat([summed, house])
        # Return #
        return df

    @property_cached
    def wide_format(self):
        """
        Transform the dataframe into the wide format
        With waste categories as column names.
        """
        # Load #
        df = self.collapse_ind
        # Pivot #
        df = df.pipe(pandas.pivot_table,
                     index   = ['country', 'year', 'nace_r2', 'wst_oper'],
                     columns = ['waste'],
                     values  = ['kg_dry'])
        # Only one level on the column index #
        df.columns = df.columns.droplevel()
        # Order the waste categories #
        df = df.reindex(columns=wastes_created)
        # Order the nace activities #
        df = df.sort_index(level=2, ascending=False)
        # Return #
        return df

    # ----------------------------- Reporting ------------------------------- #
    @property_cached
    def summary_recovered(self):
        """
        Produce a dataframe that summarizes the quantities of waste disposed
        and recovered, for every waste category, in the household and
        industrial categories.
        """
        # Load #
        df = self.collapse_ind
        # Pivot #
        df = df.pipe(pandas.pivot_table,
                     index   = ['country', 'year', 'nace_r2', 'waste'],
                     columns = ['wst_oper'],
                     values  = ['kg_dry'])
        # Only one level on the column index #
        df.columns = df.columns.droplevel()
        # Split into recovered and disposed #
        recovery = trt_names[trt_names['category'] == 'recovery']['treatment']
        disposal = trt_names[trt_names['category'] == 'disposal']['treatment']
        # Sum the treatment categories that belong #
        df['recovery'] = sum(df[trt] for trt in list(recovery))
        df['disposal'] = sum(df[trt] for trt in list(disposal))
        # Remove other columns #
        df = df.drop(columns=[trt for trt in trt_names['treatment']])
        # Add total #
        df['dry_mass'] = df.sum(axis=1)
        # Add wet mass #
        df['wet_mass'] = df['dry_mass'] / (1 - wet_coefs)
        # Remove lines that are all zeros #
        #df = df.query('dry_mass != 0.0')
        # Reorder columns #
        index = ['wet_mass', 'dry_mass', 'disposal', 'recovery']
        df = df.reindex(columns=index)
        # Return #
        return df

    # -------------------------------- Cache -------------------------------- #
    @property
    def cache_path(self):
        """Specify where on the file system we will pickle the property."""
        return cache_dir + 'dataframes/' + 'waste_spread.pickle'

###############################################################################
# Create singleton #
waste_ana = WasteAnalysis()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this class like this:

    >>> from waste_flow.spreading import spread
    >>> print(spread.by_nace)
"""

# Built-in modules #

# Internal modules #
from waste_flow.common import spread_coefs, nace_selected

# First party modules #
from plumbing.cache import property_cached

# Third party modules #

###############################################################################
class WasteSpreading:

    # ----------------------------- Properties ------------------------------ #
    @property_cached
    def by_nace(self):
        """
        Returns a dictionary where every key is a nace sector name
        and every value is a dataframe looking like this:

                  W072  W073  W075  W076  W091  W092  W093    W101
        waste
        W072_hh    1.0  0.00   0.0   0.0   0.0   0.0   0.0  0.1505
        W075_hh    0.0  0.00   1.0   0.0   0.0   0.0   0.0  0.0075
        W091_hh    0.0  0.00   0.0   0.0   1.0   0.0   0.0  0.2500
        W092_hh    0.0  0.00   0.0   0.0   0.0   1.0   0.0  0.0400
        W7376_hh   0.0  0.54   0.0   0.4   0.0   0.0   0.0  0.0024
        W9999_hh   0.0  0.00   0.0   0.0   0.0   0.0   0.0  0.0335
        """
        # Make a dataframe for every sheet in the original xls #
        result = {k: df for k, df in spread_coefs.groupby('nace')}
        # Remove the default spreading from the dict #
        default_spread = result.pop('default')
        # Apply the default to all non-mentioned nace #
        result = {k: result[k]
                     if   k in result
                     else default_spread.copy()
                     for  k in nace_selected}
        # Format each dataframe #
        result = {k: self.format_spread(df) for k, df in result.items()}
        # Return #
        return result

    # ------------------------------ Methods -------------------------------- #
    def format_spread(self, df):
        """Format a given dataframe."""
        # Drop columns #
        df = df.drop(columns=['nace'])
        # Set index #
        df = df.set_index('waste')
        # Assume zero everywhere #
        df = df.fillna(0)
        # Return #
        return df

###############################################################################
# Create singleton #
spread = WasteSpreading()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this file like this:

    >>> from waste_flow.common import waste_names
    >>> print(waste_names)
"""

# Built-in modules #

# Internal modules #
from waste_flow import module_dir

# First party modules #

# Third party modules #
import pandas

###############################################################################
# Load nace names #
nace_names = module_dir + 'extra_data_csv/nace_to_full_name.csv'
nace_names = pandas.read_csv(str(nace_names))

# Load waste names #
waste_names = module_dir + 'extra_data_csv/waste_to_full_name.csv'
waste_names = pandas.read_csv(str(waste_names))

# Load treatment names #
trt_names = module_dir + 'extra_data_csv/treatment_to_full_name.csv'
trt_names = pandas.read_csv(str(trt_names))

# Will be used to filter later #
nace_selected   = list(nace_names['nace'])
trt_selected    = list(trt_names['treatment'])
wastes_selected = list(waste_names.query("category == 'eurostat'")['waste'])
wastes_created  = list(waste_names.query("category != 'eurostat'")['waste'])

# Load spread coefficients #
spread_coefs = module_dir + 'extra_data_csv/waste_spreading.csv'
spread_coefs = pandas.read_csv(str(spread_coefs))

# Load dry to wet coefficients #
wet_coefs = waste_names.query("category != 'eurostat'")
wet_coefs = wet_coefs.reset_index(drop=True)
wet_coefs = wet_coefs.set_index('waste')['wet_fraction']

# Load country codes #
country_codes = module_dir + 'extra_data_csv/foastat_countries.csv'
country_codes = pandas.read_csv(str(country_codes))

###############################################################################
# Load names that appear in the other dataframe #
orig_w_names = set(spread_coefs['waste'])
new_w_names  = set(spread_coefs.columns) - {'waste', 'nace'}

# Check that waste names match #
assert orig_w_names == set(wastes_created)
assert new_w_names  == set(wastes_selected)

###############################################################################
# Split in two #
waste_names_ind = [w for w in waste_names['waste'] if w.endswith('_ind')]
waste_names_hh  = [w for w in waste_names['waste'] if w.endswith('_hh')]
waste_names_ind = waste_names.query("waste in @waste_names_ind")
waste_names_hh  = waste_names.query("waste in @waste_names_hh")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this class like this:

    >>> from waste_flow import outputs
    >>> print(outputs.waste_breakdown())
    >>> print(outputs.summary_recovered())
"""

# Built-in modules #

# Internal modules #
from waste_flow import cache_dir
from waste_flow.country import all_countries

# First party modules #
from plumbing.xls_tables import MultiDataFrameXLS

# Third party modules #

###############################################################################
class Outputs:
    """Takes care of creating some files that contain outputs."""

    # ----------------------------- Properties ------------------------------ #
    @property
    def path(self):
        """Specify where on the file system we will pickle the property."""
        return cache_dir + 'outputs/' + self.short_name + '.xlsx'

    # ------------------------------ Methods -------------------------------- #
    def __call__(self):
        """
        Create an excel file with one sheet per country.
        """
        # Write each country #
        sheet_to_dfs = {c.code: self.one_country(c) for c in all_countries}
        # Save #
        multi_xls = MultiDataFrameXLS(sheet_to_dfs, self.path)
        # Do it #
        return multi_xls()

    def one_country(self, country):
        """
        Create an excel sheet with every year that a
        given country has.
        """
        # Get the dataframe #
        df = getattr(country, self.df_name)
        df = df.reset_index()
        # Group by year #
        groups = df.groupby('year')
        # Initialize #
        all_dfs = []
        # Loop #
        for year, table in groups:
            # Drop the year column #
            table = table.drop('year', axis=1)
            table = table.set_index('nace_r2')
            # Table and title #
            sheet = {'dataframe': table, 'title': "YEAR %s" % year}
            # Add other optional labels #
            labels = ['x_title', 'y_title',
                      'x_label', 'y_label',
                      'x_extra', 'y_extra']
            for label in labels: sheet[label] = getattr(self, label, None)
            # Append #
            all_dfs.append(sheet)
        # Return #
        return all_dfs

###############################################################################
class WasteBreakdown(Outputs):
    short_name = "waste_breakdown"
    df_name    = "wide_format"
    x_title    = "Custom Waste Categories (values are kilograms)"
    y_title    = "Nace"

class SummaryRecovered(Outputs):
    short_name = "summary_recovered"
    df_name    = "summary_recovered"
    x_title    = "Summary (values are kilograms)"
    y_title    = "Nace"

###############################################################################
# Create singletons #
waste_breakdown   = WasteBreakdown()
summary_recovered = SummaryRecovered()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Internal modules #
from waste_flow         import cache_dir
from waste_flow.country import all_countries
from waste_flow.common  import waste_names_ind, waste_names_hh

# First party modules #
from plumbing.graphs             import Graph
from plumbing.cache              import property_cached
from plumbing.graphs.solo_legend import SoloLegend

# Third party modules #
from matplotlib import pyplot
import pandas

# Where to store the graphs #
base_dir = cache_dir + 'graphs/summary/'

###############################################################################
class SummaryGraph(Graph):
    """
    Every country has two SummaryGraph objects.
    One for industrial activities, one for household activities.
    The values indicate dry mass in kilograms.
    """

    # Basic params #
    formats    = ('pdf',)
    height     = 5
    width      = 13
    y_grid     = True

    # Labels for axes #
    x_label = 'Year'
    y_label = 'Dry mass in kilograms'

    def __init__(self, country, category):
        # Basic attributes #
        self.country    = country
        self.category   = category
        self.short_name = country.code + '_' + category
        # Nace filter #
        if self.category == 'industrial': self.nace_r2 = 'all_indus'
        if self.category == 'household':  self.nace_r2 = 'EP_HH'
        # Waste suffix #
        if self.category == 'industrial': self.waste_names = waste_names_ind
        if self.category == 'household':  self.waste_names = waste_names_hh
        # Waste filter #
        self.wastes = self.waste_names['waste']
        # Colors #
        self.colors = dict(self.waste_names.set_index('waste')['plot_color'])
        # Call parent #
        super().__init__(country, base_dir)

    @property_cached
    def df(self):
        # Load #
        df = self.country.summary_recovered
        # Filter for the current nace activity group #
        df = df.query("nace_r2 == @self.nace_r2")
        # Filter for the current waste group #
        df = df.query("waste in @self.wastes")
        # Take only the dry_mass column #
        df = df[['dry_mass']]
        # Swap index levels #
        df = df.reset_index()
        df = df.drop(columns=['nace_r2'])
        # Pivot #
        df = df.pipe(pandas.pivot_table,
                     index   = ['year'],
                     columns = ['waste'],
                     values  = ['dry_mass'])
        # Only one level on the column index #
        df.columns = df.columns.droplevel()
        # Return #
        return df

    def plot(self, **kwargs):
        # Make figure #
        fig  = pyplot.figure()
        axes = fig.add_subplot(111)
        # Plot every line #
        for waste in self.wastes:
            if waste not in self.df.columns: continue
            axes.plot(self.df[waste],
                      marker     = ".",
                      markersize = 20.0,
                      linewidth  = 5.0,
                      color      = self.colors[waste])
        # Leave some space around the graph #
        #pyplot.subplots_adjust(wspace=0.2, top=0.9, left=0.04, right=0.95, bottom=0.1)
        # Save #
        self.save_plot(**kwargs)
        # Convenience: return for display in notebooks for instance #
        return fig

###############################################################################
class SummaryLegend(SoloLegend):

    def __init__(self, category):
        # Basic attributes #
        self.category = category
        # Name #
        self.short_name = 'legend_' + category
        # Waste suffix #
        if self.category == 'industrial': self.waste_names = waste_names_ind
        if self.category == 'household':  self.waste_names = waste_names_hh
        # Index #
        self.waste_names = self.waste_names.set_index('waste')
        # Colors #
        self.name_to_color = dict(self.waste_names['plot_color'])
        # Call parent #
        super().__init__(base_dir=base_dir)

    @property_cached
    def label_to_color(self):
        """Mapping of each waste to type to colors."""
        df = self.waste_names.reset_index()
        df = df.set_index('full_name')
        return dict(df['plot_color'])

###############################################################################
# Every country has a several graphs (each graph has several subplots) #
countries_ind = {c.code: SummaryGraph(c, 'industrial') for c in all_countries}
countries_hh  = {c.code: SummaryGraph(c, 'household')  for c in all_countries}

# Create separate standalone legends #
legend_ind = SummaryLegend('industrial')
legend_hh  = SummaryLegend('household')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this class like this:

    >>> from waste_flow.country import countries
    >>> for country in countries.values():
    >>>     print('**' + country.long_name + '**')
    >>>     print(country.summary_recovered)
"""

# Built-in modules #
import importlib

# Internal modules #
from waste_flow.common    import country_codes
from waste_flow.zip_files import waste_gen as gen_orig

# First party modules #
from plumbing.cache import property_cached

# Third party modules #

###############################################################################
class Country:
    """Represents one country's dataset."""

    def __init__(self, code):
        # The reference country code #
        self.code = code

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.code)

    # ----------------------------- Properties ------------------------------ #
    @property_cached
    def long_name(self):
        """The long name of this country."""
        row = country_codes.query('nuts_zero_2010 == @self.code')
        return row.iloc[0]['country']

    @property
    def wide_format(self):
        return self.get_a_df('waste_flow.analysis.waste_ana',
                             'wide_format')

    @property
    def summary_recovered(self):
        return self.get_a_df('waste_flow.analysis.waste_ana',
                             'summary_recovered')

    @property
    def waste_gen(self):
        return self.get_a_df('waste_flow.generation.waste_gen',
                             'wide_format',
                             False)

    # ------------------------------ Methods -------------------------------- #
    def get_a_df(self, location, df_name, by_index=True):
        """
        Return rows that concern this country only for any
        given dataframe passed in the location argument.
        """
        # Split #
        module_path = '.'.join(location.split('.')[:-1])
        object_name = location.split('.')[-1]
        # Import #
        mod = importlib.import_module(module_path)
        obj = getattr(mod, object_name)
        # Load #
        df = getattr(obj, df_name)
        # Select rows for current country #
        if by_index: df = df.loc[[self.code]].copy()
        else:        df = df.query("country == '%s'" % self.code)
        # We don't need the country column anymore #
        if by_index: df = df.droplevel(0)
        else:        df = df.reset_index().drop(columns=['country'])
        # Return #
        return df

###############################################################################
# Get all possible countries #
all_codes = list(gen_orig.df.country.unique())

# Create every country object #
all_countries = [Country(code) for code in all_codes]
countries     = {c.code: c for c in all_countries}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

Typically you can use this class like this:

    >>> from waste_flow.generation import waste_gen
    >>> print(waste_gen.wide_format)
"""

# Built-in modules #

# Internal modules #
from waste_flow.zip_files import waste_gen as orig_gen

# First party modules #
from plumbing.cache import property_cached

# Third party modules #
import pandas

###############################################################################
class WasteGeneration:

    # ----------------------------- Formatting ------------------------------ #
    @property
    def filtered(self):
        """Filter the dataframe some more."""
        # Load #
        df = orig_gen.df
        # Filter for only the waste categories we are interested in #
        from waste_flow.common import wastes_selected
        df = df.query("waste in @wastes_selected")
        # Filter for only the nace categories we are interested in #
        from waste_flow.common import nace_selected
        df = df.query("nace_r2 in @nace_selected")
        # Return #
        return df

    @property_cached
    def long_format(self):
        """
        Transform the dataframe into the long format
        (and not the wide format).
        """
        # Load #
        df = self.filtered
        # Unpivot #
        df = df.melt(id_vars    = ['nace_r2', 'waste', 'country'],
                     value_name = 'tonnes',
                     var_name   = 'year')
        # Sort by country then by year etc. #
        columns_order = ['country', 'year', 'waste', 'nace_r2']
        df = df[columns_order + ['tonnes']]
        df = df.sort_values(columns_order)
        # Return #
        return df

    @property_cached
    def wide_format(self):
        """
        Transform the dataframe into the wide format
        (and not the long format).
        With waste categories as column names.
        """
        # Load #
        df = self.long_format
        # Pivot #
        df = df.pipe(pandas.pivot_table,
                     index   = ['country', 'year', 'nace_r2'],
                     columns = ['waste'],
                     values  = ['tonnes'])
        # Only one level on the column index #
        df.columns = df.columns.droplevel()
        # Return #
        return df

###############################################################################
# Create singleton #
waste_gen = WasteGeneration()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.

JRC Biomass Project.
Unit D1 Bioeconomy.

These contains the tests to be run automatically with the pytest
executable. To run all the tests just type the following on your
terminal:

    $ python3 -m pip install --upgrade --user pytest
    $ pytest
"""

# Built-in modules #

# Internal modules #

# First party modules #

# Third party modules #
import pytest

###############################################################################
# --------------------------------- Import ---------------------------------- #
def test_package_import():
    import waste_flow
    return waste_flow

###############################################################################
# -------------------------------- Download --------------------------------- #
@pytest.mark.skip(reason="Takes too long to download."
                         "Don't want to spam eurostat.")
def test_gen_download():
    from waste_flow.zip_files import waste_gen
    return waste_gen.refresh_cache()

@pytest.mark.skip(reason="Takes too long to download."
                         "Don't want to spam eurostat.")
def test_trt_download():
    from waste_flow.zip_files import waste_trt
    return waste_trt.refresh_cache()

###############################################################################
# ------------------------------- Zip files --------------------------------- #
def test_gen_parsing():
    from waste_flow.zip_files import waste_gen
    del waste_gen.processed_csv
    return waste_gen.processed_csv

def test_trt_parsing():
    from waste_flow.zip_files import waste_trt
    del waste_trt.processed_csv
    return waste_trt.processed_csv

###############################################################################
# -------------------------------- Datasets --------------------------------- #
def test_gen_dataset():
    from waste_flow.generation import waste_gen
    return waste_gen.wide_format

def test_trt_dataset():
    from waste_flow.treatment import waste_trt
    return waste_trt.normalized

###############################################################################
# -------------------------------- Analysis --------------------------------- #
def test_spread():
    from waste_flow.analysis import waste_ana
    del waste_ana.spread_waste
    return waste_ana.spread_waste

def test_summary():
    from waste_flow.analysis import waste_ana
    return waste_ana.summary_recovered